from collections import defaultdict
from contextlib import contextmanager
from dataclasses import asdict
from enum import Enum
import inspect
import json
import os
import types
import logging
from rich import print
import subprocess

from eyeball_pp.graders import model_based_grader
from .recorders import (
    ApiClientRecorder,
    Checkpoint,
    ComparisonResult,
    FileRecorder,
    MemoryRecorder,
    EvalRecorder,
    get_input_hash,
)
from .comparators import model_graded_comparator, output_feedback_from_scores
from .classes import (
    Criteria,
    FeedbackResult,
    MultiOutputFeedback,
    MultiOutputScores,
    OutputComparator,
    OutputGrader,
    OutputScore,
    OutputFeedback,
    EvaluatorConfig,
    TASK_OUTPUT_KEY,
)

import random
from statistics import variance
import threading
from typing import (
    Callable,
    Iterable,
    Iterator,
    Protocol,
    TypeVar,
    Optional,
    Any,
    runtime_checkable,
)
from functools import wraps
import datetime
from .utils import get_score_map, get_user_input, output_table, time_to_str
from .system_state import bucketize_checkpoints, get_system_tags
from tqdm import tqdm

GREEN = "\x1b[32m"
ORANGE = "\x1b[33m"
RED = "\x1b[31m"
BOLD = "\x1b[1m"
UNDERLINE = "\x1b[4m"
ITALIC = "\x1b[3m"
HEADING_BG = "\x1b[103m"
SUCCESS_CUTOFF = 0.5
END_CLR = "\x1b[0m"

logger = logging.getLogger(__name__)


@runtime_checkable
class JsonSerializable(Protocol):
    def to_json(self) -> str:
        ...


T = TypeVar("T", int, float, str, bool, bytes, dict, list, None, JsonSerializable)


class EvaluatorMode(Enum):
    RECORD = "record"
    COMPARE_CHECKPOINTS = "compare_checkpoints"
    RERUN_EXAMPLES = "rerun_examples"
    RATE_EXAMPLES = "rate_examples"


class Evaluator:
    def __init__(self, **config_kwargs) -> None:
        self.config = EvaluatorConfig()
        self.mode: EvaluatorMode = EvaluatorMode.RECORD
        self.running_in_notebook = False
        try:
            get_ipython()  # type: ignore
            self.running_in_notebook = True
            self.current_recorder_state = types.SimpleNamespace()
        except NameError:
            self.current_recorder_state = threading.local()  # type: ignore
        self.set_config(**config_kwargs)

    def _get_config(self, **override_config_kwargs) -> EvaluatorConfig:
        return EvaluatorConfig._merge(self.config, **override_config_kwargs)

    def get_recorder(self) -> EvalRecorder:
        return self.recorder
    
    def get_config_dict(self) -> dict[str, Any]:
        return asdict(self.config)

    def set_config(self, **config_kwargs) -> None:
        self.config = EvaluatorConfig._merge(self.config, **config_kwargs)
        self.data_dir = os.path.join(self.config.dir_path, "eyeball_data")

        if self.config.record_in_memory:
            self.recorder: EvalRecorder = MemoryRecorder()
        elif self.config.api_key is not None:
            self.recorder = ApiClientRecorder(
                api_key=self.config.api_key, api_url=self.config.api_url
            )
        elif api_key := os.environ.get("EYEBALLPP_API_KEY"):
            self.recorder = ApiClientRecorder(api_key=api_key, api_url=self.config.api_url)
        elif self.config.sample_rate == 0:
            self.recorder = MemoryRecorder()
        else:
            self.recorder = FileRecorder(self.data_dir) 

    @contextmanager
    def start_recording_session(
        self,
        task_name: str,
        checkpoint_id: Optional[str] = None,
        checkpoint_id_to_rerun: Optional[str] = None,
    ) -> Iterator[None]:
        if not self.running_in_notebook and hasattr(
            self.current_recorder_state, "recorder_checkpoint_id"
        ):
            if self.mode == EvaluatorMode.RERUN_EXAMPLES:
                yield None
                return

            # This should not happen but if it does, we should not overwrite the previous checkpoint id
            # Instead we should set it to None so there is no confusion
            # If the user calls get_eval_params this will raise an error
            self.current_recorder_state.task_name = None
            self.current_recorder_state.recorder_checkpoint_id = None
            self.current_recorder_state.checkpoint_id_to_rerun = None
            yield None
            return

        self.current_recorder_state.task_name = task_name
        if checkpoint_id is None:
            checkpoint_id = datetime.datetime.utcnow().isoformat()

        self.current_recorder_state.recorder_checkpoint_id = checkpoint_id
        self.current_recorder_state.checkpoint_id_to_rerun = checkpoint_id_to_rerun
        yield None

        for tag in get_system_tags():
            self.recorder.add_checkpoint_tag(
                task_name=task_name,
                checkpoint_id=checkpoint_id,
                tag=tag,
            )

        del self.current_recorder_state.task_name
        del self.current_recorder_state.recorder_checkpoint_id

    def _should_record(self, **config_kwargs) -> bool:
        if self.mode == EvaluatorMode.COMPARE_CHECKPOINTS:
            return False

        if self.mode == EvaluatorMode.RERUN_EXAMPLES:
            return True

        config = self._get_config(**config_kwargs)
        if config.sample_rate == 0:
            return False
        return config.sample_rate == 1.0 or random.random() < config.sample_rate

    def record_input(
        self,
        variable_name: str,
        value: T,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
        **config_kwargs,
    ) -> None:
        if self._should_record(**config_kwargs):
            task_name, checkpoint_id = self._get_recorder_state(
                task_name, checkpoint_id
            )

            serialized_value = self._serialize(value)
            self.recorder.record_input_variable(
                task_name=task_name,
                checkpoint_id=checkpoint_id,
                variable_name=variable_name,
                value=serialized_value,
            )

    def record_intermediary_state(
        self,
        state_name: str,
        state_value: str,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
        **config_kwargs,
    ) -> None:
        if self._should_record(**config_kwargs):
            task_name, checkpoint_id = self._get_recorder_state(
                task_name, checkpoint_id
            )

            self.recorder.record_intermediary_state(
                task_name=task_name,
                checkpoint_id=checkpoint_id,
                state_name=state_name,
                state_value=state_value,
            )

    def record_output(
        self,
        value: Any,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
        **config_kwargs,
    ) -> None:
        if self._should_record(**config_kwargs):
            task_name, checkpoint_id = self._get_recorder_state(
                task_name, checkpoint_id
            )
            self.recorder.record_output(
                task_name=task_name,
                checkpoint_id=checkpoint_id,
                output=self._serialize(value),
            )

    def _get_last_checkpoint(
        self,
        task_name: str,
        input_hash: str,
        current_checkpoint_id: str,
    ) -> Optional[Checkpoint]:
        checkpoints = self.recorder.get_latest_checkpoints(
            task_name, input_hash, num_checkpoints=2
        )
        if len(checkpoints) == 0:
            return None

        if checkpoints[0].checkpoint_id == current_checkpoint_id:
            if len(checkpoints) == 1:
                return None
            return checkpoints[1]

        else:
            return checkpoints[0]

    def _get_recorder_state(
        self,
        task_name: Optional[str] = None,
        recorder_checkpoint_id: Optional[str] = None,
    ) -> tuple[str, str]:
        if task_name is None and hasattr(self.current_recorder_state, "task_name"):
            task_name = self.current_recorder_state.task_name
        if task_name is None:
            raise ValueError(
                "Must provide a task name or start a recording session to record a variable"
            )

        if recorder_checkpoint_id is None and hasattr(
            self.current_recorder_state, "recorder_checkpoint_id"
        ):
            recorder_checkpoint_id = self.current_recorder_state.recorder_checkpoint_id
        if recorder_checkpoint_id is None:
            raise ValueError(
                "Must provide a checkpoint id or start a recording session to record a variable"
            )

        return task_name, recorder_checkpoint_id

    def _get_recorded_task_name(self, task_name: Optional[str]) -> str:
        if task_name is not None:
            return task_name
        if hasattr(self.current_recorder_state, "task_name"):
            return self.current_recorder_state.task_name

        task_names = self.recorder.get_task_names()
        if len(task_names) == 0:
            raise ValueError("No task names found")
        elif len(task_names) == 1:
            return task_names[0]
        else:
            raise ValueError(
                f"Must provide a task_name as you have multiple tasks recorded -- Found: {task_names}"
            )

    def get_eval_params(
        self,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ) -> Optional[dict[str, Any]]:
        task_name, checkpoint_id = self._get_recorder_state(task_name, checkpoint_id)

        checkpoint = self.recorder.get_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        if checkpoint is None:
            return None
        return checkpoint.eval_params

    def get_eval_param(
        self,
        param_name: str,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ) -> Optional[Any]:
        params = self.get_eval_params(task_name=task_name, checkpoint_id=checkpoint_id)
        if params is None:
            return None
        return params.get(param_name)

    def _serialize(self, value: T) -> str:
        if isinstance(value, JsonSerializable):
            return value.to_json()
        else:
            return json.dumps(value)

    def record_task(
        self,
        args_to_record: list[str],
        task_name: Optional[str] = None,
        checkpoint_id_arg_name: Optional[str] = None,
        eval_params: Optional[dict[str, Any]] = None,
        request_user_feedback_probability: float = 0.0,
        **config_kwargs,
    ) -> Callable[..., Callable]:
        def _decorator(fn: Callable[..., T]) -> Callable[..., T]:
            @wraps(fn)
            def _wrapper(*args, **kwargs) -> Any:
                if not self._should_record(**config_kwargs):
                    return fn(*args, **kwargs)

                (
                    fn_arg_names,
                    _,
                    _,
                    _,
                    _,
                    _,
                    _,
                ) = inspect.getfullargspec(fn)

                # Set task name
                if self.mode == EvaluatorMode.RERUN_EXAMPLES:
                    local_task_name = self.current_recorder_state.task_name
                else:
                    local_task_name = task_name or fn.__name__

                # Set checkpoint id
                if self.mode == EvaluatorMode.RERUN_EXAMPLES:
                    recorder_checkpoint_id = (
                        self.current_recorder_state.recorder_checkpoint_id
                    )
                elif checkpoint_id_arg_name is not None:
                    if checkpoint_id_arg_name in kwargs:
                        recorder_checkpoint_id = kwargs[checkpoint_id_arg_name]
                    else:
                        try:
                            checkpoint_id_arg_index = fn_arg_names.index(
                                checkpoint_id_arg_name
                            )
                            recorder_checkpoint_id = args[checkpoint_id_arg_index]
                        except ValueError:
                            recorder_checkpoint_id = (
                                datetime.datetime.utcnow().isoformat()
                            )
                else:
                    recorder_checkpoint_id = datetime.datetime.utcnow().isoformat()

                # Set params
                if self.mode == EvaluatorMode.RERUN_EXAMPLES:
                    params = self.get_eval_params(
                        task_name=local_task_name,
                        checkpoint_id=recorder_checkpoint_id,
                    )
                    # TODO: maybe we should not override with default params in this mode. Can lead to some confusion
                    if (params is None or len(params) == 0) and eval_params is not None:
                        self.recorder.record_eval_params(
                            task_name=local_task_name,
                            checkpoint_id=recorder_checkpoint_id,
                            eval_params=eval_params,
                        )
                elif eval_params is not None:
                    self.recorder.record_eval_params(
                        task_name=local_task_name,
                        checkpoint_id=recorder_checkpoint_id,
                        eval_params=eval_params,
                    )

                for arg_name, arg_val in zip(fn_arg_names, args):
                    if arg_name in args_to_record:
                        self.record_input(
                            task_name=local_task_name,
                            checkpoint_id=recorder_checkpoint_id,
                            variable_name=arg_name,
                            value=arg_val,
                        )

                for kwarg_name, kwarg_val in kwargs.items():
                    if kwarg_name in args_to_record:
                        self.record_input(
                            task_name=local_task_name,
                            checkpoint_id=recorder_checkpoint_id,
                            variable_name=kwarg_name,
                            value=kwarg_val,
                        )

                if self.mode == EvaluatorMode.RERUN_EXAMPLES:
                    result = fn(*args, **kwargs)  # type: ignore
                else:
                    with self.start_recording_session(
                        task_name=local_task_name,
                        checkpoint_id=recorder_checkpoint_id,
                    ):
                        result = fn(*args, **kwargs)  # type: ignore

                self.record_output(
                    task_name=local_task_name,
                    checkpoint_id=recorder_checkpoint_id,
                    value=result,
                )
                if request_user_feedback_probability > 0:
                    if random.random() < request_user_feedback_probability:
                        print("Requesting user feedback")
                        self.request_user_feedback(
                            output=str(result),
                            task_name=local_task_name,
                            checkpoint_id=recorder_checkpoint_id,
                        )
                return result

            return _wrapper

        return _decorator

    def request_user_feedback(
        self,
        output: str,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
        print_inputs: bool = False,
    ) -> Optional[OutputFeedback]:
        task_name, checkpoint_id = self._get_recorder_state(task_name, checkpoint_id)
        checkpoint = self.recorder.get_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        if checkpoint is None:
            return None

        if print_inputs:
            print(checkpoint.get_input_var_str())

        feedback_result = FeedbackResult[
            get_user_input(
                f"What do you think of the result?\n{output}\n",
                choices=[f.name for f in FeedbackResult],
            )
        ]
        details = get_user_input(
            "Any details about your feedback (optional)\n", choices=None
        )

        output_feedback = OutputFeedback(result=feedback_result, message=details)

        self.recorder.record_output_feedback(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            feedback=MultiOutputFeedback({TASK_OUTPUT_KEY: output_feedback}),
        )
        return output_feedback

    def rerun_recorded_examples(
        self,
        *eval_params_list: dict[str, Any],
        task_name: Optional[str] = None,
        input_hashes: Optional[list[str]] = None,
        limit: Optional[int] = None,
        randomize: Optional[bool] = False,
    ) -> Iterable[dict[str, Any]]:
        if task_name is None:
            task_names = self.recorder.get_task_names()
            if len(task_names) == 0:
                raise ValueError("No task names found")
            elif len(task_names) == 1:
                task_name = task_names[0]
            else:
                raise ValueError(
                    f"Must provide a task_name as you have multiple tasks recorded -- Found: {task_names}"
                )

        if input_hashes is None or len(input_hashes) == 0:
            input_hashes = self.recorder.get_input_hashes(task_name=task_name)

        if randomize and input_hashes:
            random.shuffle(input_hashes)

        if limit is not None:
            input_hashes = input_hashes[:limit]

        try:
            self.mode = EvaluatorMode.RERUN_EXAMPLES
            rerun_metadata = {
                "id": datetime.datetime.utcnow().isoformat(),
            }

            print(f"Will rerun {len(input_hashes)} inputs for task:`{task_name}`")

            for idx, input_hash in tqdm(enumerate(input_hashes)):
                recorder_checkpoint_id = datetime.datetime.utcnow().isoformat()
                last_checkpoint = self._get_last_checkpoint(
                    task_name=task_name,
                    input_hash=input_hash,
                    current_checkpoint_id=recorder_checkpoint_id,
                )
                if last_checkpoint is None:
                    continue

                self.checkpoint_id_to_rerun = last_checkpoint.checkpoint_id
                checkpoint_to_rerun = last_checkpoint
                if checkpoint_to_rerun is None:
                    continue

                input_vars = {
                    k: json.loads(v)
                    for k, v in checkpoint_to_rerun.get_input_variables().items()
                }
                logger.debug(
                    f"\nRerunning input #{idx}:\n{checkpoint_to_rerun.get_input_var_str()}"
                )

                if len(eval_params_list) > 0:
                    for eval_params in eval_params_list:
                        with self.start_recording_session(
                            task_name=task_name,
                            checkpoint_id=recorder_checkpoint_id,
                            checkpoint_id_to_rerun=checkpoint_to_rerun.checkpoint_id,
                        ):
                            self.recorder.record_eval_params(
                                task_name=task_name,
                                checkpoint_id=recorder_checkpoint_id,
                                eval_params=eval_params,
                                rerun_metadata=rerun_metadata,
                            )
                            logger.debug(f"Using eval params: {eval_params}")
                            yield input_vars
                else:
                    with self.start_recording_session(
                        task_name=task_name,
                        checkpoint_id=recorder_checkpoint_id,
                        checkpoint_id_to_rerun=checkpoint_to_rerun.checkpoint_id,
                    ):
                        self.recorder.record_eval_params(
                            task_name=task_name,
                            checkpoint_id=recorder_checkpoint_id,
                            eval_params={},
                            rerun_metadata=rerun_metadata,
                        )
                        yield input_vars
        finally:
            self.mode = EvaluatorMode.RECORD

    def rate_recorded_examples(
        self,
        *input_hashes: str,
        task_name: Optional[str] = None,
    ) -> None:
        if task_name is None:
            task_names = self.recorder.get_task_names()
            if len(task_names) == 0:
                raise ValueError("No task names found")
            elif len(task_names) == 1:
                task_name = task_names[0]
            else:
                raise ValueError(
                    f"Must provide a task_name as you have multiple tasks recorded -- Found: {task_names}"
                )

        input_hashes_lst: list[str] = list(input_hashes)

        if len(input_hashes_lst) == 0:
            input_hashes_lst = self.recorder.get_input_hashes(task_name=task_name)

        try:
            self.mode = EvaluatorMode.RATE_EXAMPLES
            for input_hash in input_hashes_lst:
                already_seen_outputs_to_feedback: dict[str, MultiOutputFeedback] = {}
                checkpoints = self.recorder.get_latest_checkpoints(
                    task_name, input_hash, num_checkpoints=4
                )

                if len(checkpoints) == 0:
                    continue

                print(f"\nFor inputs:\n{checkpoints[0].get_input_var_str()}")
                for checkpoint in checkpoints:
                    if checkpoint.output is None:
                        continue

                    if not checkpoint.feedback:
                        if feedback := already_seen_outputs_to_feedback.get(
                            checkpoint.output
                        ):
                            self.recorder.record_output_feedback(
                                task_name=task_name,
                                checkpoint_id=checkpoint.checkpoint_id,
                                feedback=feedback,
                            )
                        else:
                            output_feedback = self.request_user_feedback(
                                checkpoint.output or "",
                                task_name=task_name,
                                checkpoint_id=checkpoint.checkpoint_id,
                            )
                            if output_feedback is not None:
                                already_seen_outputs_to_feedback[
                                    checkpoint.output
                                ] = MultiOutputFeedback(
                                    {TASK_OUTPUT_KEY: output_feedback}
                                )
                    elif new_feedback := already_seen_outputs_to_feedback.get(
                        checkpoint.output
                    ):
                        if (
                            new_feedback[TASK_OUTPUT_KEY].result
                            != checkpoint.feedback[TASK_OUTPUT_KEY].result
                        ):
                            print(
                                f"For output: {checkpoint.output}\nOld feedback {checkpoint.feedback[TASK_OUTPUT_KEY]} is different from new feedback {new_feedback[TASK_OUTPUT_KEY]}"
                            )
                            print(f"Updating feedback to {new_feedback}")
                            self.recorder.record_output_feedback(
                                task_name=task_name,
                                checkpoint_id=checkpoint.checkpoint_id,
                                feedback=new_feedback,
                            )
                    else:
                        print(
                            f"Output: {checkpoint.output[:140]} already has feedback: {checkpoint.feedback.get(TASK_OUTPUT_KEY)}"
                        )
        finally:
            self.mode = EvaluatorMode.RECORD

    def record_human_feedback(
        self,
        feedback: OutputFeedback,
        details: Optional[str] = None,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ):
        task_name, checkpoint_id = self._get_recorder_state(task_name, checkpoint_id)
        pass

    def delete_checkpoints_for_input_vars(
        self,
        input_vars: dict[str, str],
        task_name: Optional[str] = None,
    ) -> None:
        if task_name is None:
            task_names = self.recorder.get_task_names()
            if len(task_names) == 0:
                raise ValueError("No task names found")
            elif len(task_names) == 1:
                task_name = task_names[0]
            else:
                raise ValueError(
                    f"Must provide a task_name as you have multiple tasks recorded -- Found: {task_names}"
                )
        input_hash = get_input_hash(
            {k: self._serialize(v) for k, v in input_vars.items()}
        )
        self.recorder.delete_checkpoints_for_input_hash(task_name, input_hash)

    def evaluate_system(
        self,
        task_name: Optional[str] = None,
        num_input_hashes: int = 5,
        num_checkpoints_per_input_hash: int = 3,
        task_objective: Optional[str] = None,
        grading_criteria: Optional[list[Criteria]] = None,
        grading_criteria_custom: Optional[dict[str, str]] = None,
        output_comparator: Optional[OutputComparator] = None,
        output_grader: Optional[OutputGrader] = None,
        intermediate_objectives: Optional[dict[str, str]] = None,
        use_cached_scores: bool = True,
        use_cached_comparisons: bool = True,
    ) -> None:
        task_name = self._get_recorded_task_name(task_name)

        if output_grader is None:
            output_grader = model_based_grader

        input_hashes_list = self.recorder.get_input_hashes(task_name)
        random.shuffle(input_hashes_list)
        input_hashes_list = input_hashes_list[:num_input_hashes]

        print(f"Evaluating {len(input_hashes_list)} inputs for task:`{task_name}`")
        try:
            self.mode = EvaluatorMode.COMPARE_CHECKPOINTS

            num_comparisons = 0
            for idx, input_hash in enumerate(
                tqdm(
                    input_hashes_list,
                    bar_format="{l_bar}{bar}|{n_fmt}/{total_fmt}",
                )
            ):
                checkpoints_list = self.recorder.get_latest_checkpoints(
                    task_name,
                    input_hash,
                    num_checkpoints=num_checkpoints_per_input_hash,
                )

                checkpoints: dict[str, Checkpoint] = {
                    c.checkpoint_id: c
                    for c in checkpoints_list
                    if c.output is not None
                    and c.output != "null"
                    and c.output != "None"
                }
                checkpoint_ids = sorted(checkpoints.keys(), reverse=True)

                if len(checkpoint_ids) == 0:
                    continue

                if output_grader is not None:
                    # TODO: change output scorer to score multiple output types
                    logger.debug(
                        f"\nInput #{idx} - Grading {len(checkpoint_ids)} checkpoints"
                    )
                    for checkpoint_id in tqdm(
                        checkpoint_ids,
                        desc=f"Checkpoints for Input #{idx}",
                        bar_format="{l_bar}{bar}|{n_fmt}/{total_fmt}",
                        leave=False,
                    ):
                        checkpoint_to_score = checkpoints[checkpoint_id]
                        if (
                            use_cached_scores
                            and checkpoint_to_score.scores
                            and TASK_OUTPUT_KEY in checkpoint_to_score.scores
                        ):
                            logger.debug(
                                f"Using cached score for {checkpoint_id}: {checkpoint_to_score.scores[TASK_OUTPUT_KEY].score}"
                            )
                            continue

                        assert checkpoint_to_score.output is not None
                        output_score = output_grader(
                            input_variables=checkpoint_to_score.get_input_variables(),
                            output=checkpoint_to_score.output,
                            intermediary_state=checkpoint_to_score.intermediary_state,
                            objective=task_objective,
                            criteria=grading_criteria,
                            custom_criteria=grading_criteria_custom,
                        )
                        multi_output_scores = MultiOutputScores(
                            {TASK_OUTPUT_KEY: output_score}
                        )
                        self.recorder.record_output_scores(
                            task_name=task_name,
                            checkpoint_id=checkpoint_id,
                            scores=multi_output_scores,
                        )
                        checkpoint_to_score.scores = multi_output_scores
                        logger.debug(f"Scored {checkpoint_id}: {output_score.score}")

                if len(checkpoint_ids) < 2:
                    continue

                # Newer checkpoints are at the start of the list so for every comparison
                # we compare the newer_checkpoint_id with the older_checkpoint_id
                to_compare = [
                    (checkpoint_ids[i], checkpoint_ids[i + 1])
                    for i in range(len(checkpoint_ids) - 1)
                ]

                logger.debug(
                    f"\nInput #{idx} - Running {len(to_compare)} comparison(s)"
                )

                for newer_checkpoint_id, older_checkpoint_id in tqdm(
                    to_compare, desc="Comparisons", disable=True
                ):
                    newer_checkpoint = self.recorder.get_checkpoint(
                        task_name=task_name,
                        checkpoint_id=newer_checkpoint_id,
                    )
                    older_checkpoint = self.recorder.get_checkpoint(
                        task_name=task_name,
                        checkpoint_id=older_checkpoint_id,
                    )

                    if newer_checkpoint is None or older_checkpoint is None:
                        print(
                            f"Could not find checkpoint {newer_checkpoint_id} or {older_checkpoint_id}"
                        )
                        continue

                    newer_checkpoint_output = newer_checkpoint.output
                    older_checkpoint_output = older_checkpoint.output

                    assert newer_checkpoint_output is not None
                    assert older_checkpoint_output is not None

                    should_record_comparison = True
                    if use_cached_comparisons and (
                        comparator_result := self.recorder.get_comparison_result(
                            task_name=task_name,
                            older_checkpoint_id=older_checkpoint_id,
                            newer_checkpoint_id=newer_checkpoint_id,
                        )
                    ):
                        logger.debug(f"Using cached comparison result for {input_hash}")
                        should_record_comparison = False
                        comparison_feedback = comparator_result.feedback
                    elif output_comparator is not None:
                        comparison_feedback = output_comparator(
                            input_variables=newer_checkpoint.get_input_variables(),
                            older_checkpoint_output=older_checkpoint_output,
                            newer_checkpoint_output=newer_checkpoint_output,
                            task_objective=task_objective or "",
                            objectives_intermediary_state=intermediate_objectives,
                            older_checkpoint_intermediary_state=older_checkpoint.intermediary_state,
                            newer_checkpoint_intermediary_state=newer_checkpoint.intermediary_state,
                        )
                    elif output_grader is not None:
                        comparison_feedback = output_feedback_from_scores(
                            older_scores=older_checkpoint.scores,
                            newer_scores=newer_checkpoint.scores,
                        )
                    else:
                        raise ValueError(
                            "Should not happen. We need an output comparator or output scorer"
                        )

                    new_unique_params = []
                    old_unqiue_params = []
                    for key in (
                        newer_checkpoint.eval_params.keys()
                        | older_checkpoint.eval_params.keys()
                    ):
                        param_val_a = newer_checkpoint.eval_params.get(key)
                        param_val_b = older_checkpoint.eval_params.get(key)
                        if param_val_a != param_val_b:
                            if param_val_a is not None:
                                new_unique_params.append(f"{key}={param_val_a}")
                            elif param_val_b is not None:
                                old_unqiue_params.append(f"{key}={param_val_b}")

                    new_unique_params_str = (
                        ("(" + ", ".join(new_unique_params) + ")")
                        if len(new_unique_params) > 0
                        else ""
                    )
                    old_unique_params_str = (
                        ("(" + ", ".join(old_unqiue_params) + ")")
                        if len(old_unqiue_params) > 0
                        else ""
                    )

                    num_comparisons += 1
                    output_comparison_feedback = comparison_feedback[TASK_OUTPUT_KEY]
                    if output_comparison_feedback.result == FeedbackResult.NEUTRAL:
                        logger.debug(
                            f"{ORANGE}[neutral] task output is the same between checkpoints {older_checkpoint_id} {old_unique_params_str} & {newer_checkpoint_id} {new_unique_params_str} {END_CLR}"
                        )
                    elif output_comparison_feedback.result == FeedbackResult.NEGATIVE:
                        logger.debug(
                            f"{RED}[regression] task output was better in the older checkpoint {older_checkpoint_id} {old_unique_params_str} than {newer_checkpoint_id} {new_unique_params_str} {END_CLR}"
                        )
                    else:
                        logger.debug(
                            f"{GREEN}[improvement] task output improved from checkpoint {older_checkpoint_id} {old_unique_params_str} to {newer_checkpoint_id} {new_unique_params_str}{END_CLR}"
                        )
                    if should_record_comparison:
                        self.recorder.record_comparison_result(
                            task_name=task_name,
                            input_hash=input_hash,
                            result=ComparisonResult(
                                older_checkpoint_id=older_checkpoint_id,
                                newer_checkpoint_id=newer_checkpoint_id,
                                feedback=comparison_feedback,
                            ),
                        )

            self.calculate_system_health(task_name=task_name)
            self.show_data_via_streamlit(task_name=task_name)
        finally:
            self.mode = EvaluatorMode.RECORD

    def show_data_via_streamlit(self, task_name: str) -> None:
        "Run streamlit server in a subprocess"
        server_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "server.py")
        print(server_file)
        config_dict = json.dumps(self.get_config_dict())
        subprocess.Popen(
            f"python -m streamlit run {server_file} -- --task_name={task_name} --eyeball_config='{config_dict}'",
            shell=True,
        )


    def calculate_system_health(
        self,
        num_samples: int = 10,
        plotting_frequency_in_days: int = 1,
        task_name: Optional[str] = None,
    ) -> None:
        task_name = self._get_recorded_task_name(task_name=task_name)
        output_names_to_score: set[str] = set([TASK_OUTPUT_KEY])

        # calculate rolling average if we have output scores for the last num_samples checkpoints
        input_hashes = self.recorder.get_input_hashes(task_name=task_name)
        if len(input_hashes) == 0:
            print("No input hashes exist for this task")

        all_checkpoints: list[Checkpoint] = []
        input_hash_to_checkpoints: dict[str, list[Checkpoint]] = {}
        for input_hash in input_hashes:
            checkpoints = self.recorder.get_latest_checkpoints(
                task_name=task_name,
                input_hash=input_hash,
                num_checkpoints=num_samples,
            )
            input_hash_to_checkpoints[input_hash] = checkpoints
            all_checkpoints += checkpoints

        all_checkpoints_dict: dict[str, Checkpoint] = {
            c.checkpoint_id: c for c in all_checkpoints
        }

        if len(all_checkpoints) == 0:
            return

        # If the outputs have scores, we can calculate a rolling average
        scored_checkpoints: list[Checkpoint] = []
        rerunid_to_checkpoint_feedback: dict[
            str, dict[str, dict[str, Optional[FeedbackResult]]]
        ] = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: None)))

        for c in all_checkpoints:
            scored_outputs = set(c.scores.keys())
            if len(scored_outputs) > 0:
                scored_checkpoints.append(c)
                output_names_to_score |= scored_outputs

        if len(scored_checkpoints) == 0:
            # If the outputs don't have scores, we can give the better output a score of 1 and the worse output a score of 0.5 and propagate that score to the older checkpoints
            comparison_results: list[ComparisonResult] = []
            for input_hash in input_hashes:
                comparison_results += (
                    self.recorder.get_comparison_results_for_input_hash(
                        task_name=task_name,
                        input_hash=input_hash,
                        num_results=num_samples,
                    )
                )
            for cr in comparison_results:
                print(f"Feedback: {cr.feedback.keys()}")
                output_names_to_score |= set(cr.feedback.keys())
            for output_name in output_names_to_score:
                edges: dict[str, dict[str, float]] = defaultdict(lambda: dict())
                for comparison_result in comparison_results:
                    feedback = comparison_result.feedback.get(output_name)
                    if feedback is None:
                        continue

                    if (
                        new_checkpoint := all_checkpoints_dict.get(
                            comparison_result.newer_checkpoint_id
                        )
                    ) is not None:
                        if new_checkpoint.rerun_metadata:
                            rerunid_to_checkpoint_feedback[
                                new_checkpoint.rerun_metadata["id"]
                            ][comparison_result.newer_checkpoint_id][
                                output_name
                            ] = feedback.result

                    if feedback.result == FeedbackResult.POSITIVE:
                        edges[comparison_result.older_checkpoint_id][
                            comparison_result.newer_checkpoint_id
                        ] = 1.0
                    elif feedback.result == FeedbackResult.NEGATIVE:
                        edges[comparison_result.newer_checkpoint_id][
                            comparison_result.older_checkpoint_id
                        ] = 1.0
                    else:
                        edges[comparison_result.newer_checkpoint_id][
                            comparison_result.older_checkpoint_id
                        ] = 0.0
                        edges[comparison_result.older_checkpoint_id][
                            comparison_result.newer_checkpoint_id
                        ] = 0.0
                scores = get_score_map(edges)
                denom = max(scores.values()) if len(scores) > 0 else 0.0
                for checkpoint_id, score in scores.items():
                    # TODO: fetch checkpoint if it doesn't exist
                    # TODO: change input hash to checkpoints dict too
                    if checkpoint := all_checkpoints_dict.get(checkpoint_id):
                        checkpoint.scores[output_name] = OutputScore(
                            score=(score / denom) if denom > 0 else 0.0,
                            message="",
                        )

        scored_checkpoints = [c for c in all_checkpoints if len(c.scores) > 0]

        if len(scored_checkpoints) == 0:
            print(f"Not enough checkpoints with comparisons to calculate system health")
            return
        scored_checkpoints.sort(key=lambda x: x.checkpoint_id, reverse=True)
        system_health_by_date: list[dict[str, Any]] = []
        date_to_use = datetime.datetime.utcnow().date()

        system_health_by_date = []
        while scored_checkpoints[-1].created_at.date() <= date_to_use:
            num_successes = 0.0
            num_checkpoints_used = 0
            input_hash_set = set()
            for checkpoint in scored_checkpoints:
                if num_checkpoints_used >= num_samples:
                    break

                if checkpoint.created_at.date() <= date_to_use:
                    if (
                        checkpoint.scores is not None
                        and TASK_OUTPUT_KEY in checkpoint.scores
                    ):
                        if checkpoint.scores[TASK_OUTPUT_KEY].score > SUCCESS_CUTOFF:
                            num_successes += 1
                        num_checkpoints_used += 1
                        input_hash_set.add(checkpoint.get_input_hash())

            system_health_by_date.append(
                {
                    "Date": time_to_str(date_to_use),
                    "Results": f"{float(num_successes) / float(num_checkpoints_used) * 100.0: .1f}% success ({num_successes}/{num_checkpoints_used})",
                    "Stats": f"{num_checkpoints_used} datapoints, {len(input_hash_set)} unique inputs",
                }
            )
            date_to_use -= datetime.timedelta(days=plotting_frequency_in_days)

        output_table(
            system_health_by_date,
            title=f"System health for task: '{task_name}' (by Date)",
            markdown_file=os.path.join(
                self.data_dir, task_name, "system_health_by_date.md"
            ),
        )
        buckets_to_checkpoints = bucketize_checkpoints(scored_checkpoints)
        system_health_by_run_history = []
        for bucket, checkpoints in buckets_to_checkpoints.items():
            row = {"Run": str(bucket)}

            for output_name in output_names_to_score:
                num_checkpoints_used = 0
                num_successes = 0
                params_used: set[str] = set()
                input_hash_to_score: dict[str, list[float]] = {}
                for checkpoint in checkpoints:
                    if output_name in checkpoint.scores:
                        if checkpoint.scores[output_name].score > SUCCESS_CUTOFF:
                            num_successes += 1
                        num_checkpoints_used += 1
                        input_hash = checkpoint.get_input_hash()
                        if input_hash not in input_hash_to_score:
                            input_hash_to_score[input_hash] = []
                        input_hash_to_score[input_hash].append(
                            checkpoint.scores[output_name].score
                        )
                        if checkpoint.eval_params:
                            keys = sorted(checkpoint.eval_params.keys())
                            params_used.add(
                                "\n".join(
                                    f"{k}={checkpoint.eval_params[k]}" for k in keys
                                )
                            )
                if num_checkpoints_used > 0:
                    percent = float(num_successes) / float(num_checkpoints_used) * 100.0
                    column_name = (
                        "Results"
                        if output_name == TASK_OUTPUT_KEY
                        else f"{output_name}"
                    )
                    row[
                        column_name
                    ] = f"{percent: .1f}% success ({num_successes}/{num_checkpoints_used})"

                    if output_name == TASK_OUTPUT_KEY:
                        stats = f"{num_checkpoints_used} datapoints, {len(input_hash_to_score)} unique inputs"
                        row["Stats"] = stats
                        total_variance = 0.0
                        num_inputs_with_variance = 0
                        for input_hash, score_list in input_hash_to_score.items():
                            if len(score_list) > 1:
                                total_variance += variance(score_list)
                                num_inputs_with_variance += 1
                        if num_inputs_with_variance > 0:
                            row[
                                "Output Variance (higher value ⇒ unpredictable system)"
                            ] = str(total_variance / float(num_inputs_with_variance))
                        if len(params_used) == 1:
                            row["Params"] = params_used.pop()
            if len(row) > 1:
                system_health_by_run_history.append(row)
        output_table(
            system_health_by_run_history,
            title=f"System health for task: '{task_name}' (by Run History)",
            markdown_file=os.path.join(
                self.data_dir, task_name, "system_health_by_run_history.md"
            ),
        )

        # Group by reruns if they exist
        rerun_ids_to_score: dict[str, list[Checkpoint]] = defaultdict(list)
        for checkpoint in scored_checkpoints:
            if rerun_id := checkpoint.rerun_metadata.get("id"):
                rerun_ids_to_score[rerun_id].append(checkpoint)

        # rerun_rows = []
        # for rerun_id, checkpoints in sorted(
        #     rerun_ids_to_score.items(), key=lambda x: x[0], reverse=True
        # ):
        #     # We want to show how many inputs performed better in this re-run ideally
        #     # For now let's show score and then change it up
        #     row: dict[str, Any] = {"rerun_id": rerun_id}

        #     # num_better, num_worse, num_same
        #     output_feedback_status: dict[str, list[float]] = defaultdict(
        #         lambda: [0.0, 0.0, 0.0, 0.0]
        #     )
        #     all_output_names: set[str] = set()
        #     rerun_input_hashes: set[str] = set()
        #     for checkpoint in checkpoints:
        #         rerun_input_hashes.add(checkpoint.get_input_hash())
        #         for output_name, output_score in checkpoint.scores.items():
        #             all_output_names.add(output_name)
        #             if (
        #                 feedback_result := rerunid_to_checkpoint_feedback[rerun_id][
        #                     checkpoint.checkpoint_id
        #                 ]
        #             ) is not None:
        #                 if feedback_result == FeedbackResult.POSITIVE:
        #                     output_feedback_status[output_name][0] += 1.0
        #                 elif feedback_result == FeedbackResult.NEGATIVE:
        #                     output_feedback_status[output_name][1] += 1.0
        #                 else:
        #                     output_feedback_status[output_name][2] += 1.0

        #             output_feedback_status[output_name][3] += output_score.score

        #     for output_name in all_output_names:
        #         num_better, num_worse, num_same, score = output_feedback_status[
        #             output_name
        #         ]
        #         total = int(num_better + num_worse + num_same)
        #         trend_msg = ""
        #         if num_better > 0:
        #             trend_msg += f"{int(num_better)}/{total} got better\n"
        #         if num_worse > 0:
        #             trend_msg += f"{int(num_worse)}/{total} got worse\n"
        #         if num_same > 0:
        #             trend_msg += f"{int(num_same)}/{total} stayed the same\n"

        #         if trend_msg:
        #             row[f"{output_name} trend"] = trend_msg
        #         else:
        #             row[f"{output_name} relative score"] = score

        #     row["num_checkpoints_used"] = len(checkpoints)
        #     row["input_diversity"] = len(input_hashes)

        #     rerun_rows.append(row)
        # output_table(rerun_rows, title="Rerun stats")

        for output_name in output_names_to_score:
            input_specific_rows: list[dict[str, str]] = []
            for input_hash, checkpoints in input_hash_to_checkpoints.items():
                best_checkpoint: Optional[Checkpoint] = None
                most_recent_checkpoint: Optional[Checkpoint] = None
                worst_checkpoint: Optional[Checkpoint] = None

                for checkpoint in checkpoints:
                    if output_name not in checkpoint.scores:
                        continue

                    if (
                        best_checkpoint is None
                        or checkpoint.scores[output_name].score
                        > best_checkpoint.scores[output_name].score
                    ):
                        best_checkpoint = checkpoint
                    if (
                        most_recent_checkpoint is None
                        or checkpoint.created_at > most_recent_checkpoint.created_at
                    ):
                        most_recent_checkpoint = checkpoint
                    if (
                        worst_checkpoint is None
                        or checkpoint.scores[output_name].score
                        < worst_checkpoint.scores[output_name].score
                    ):
                        worst_checkpoint = checkpoint

                row_data = dict(checkpoint.input_variables)
                if best_checkpoint is not None:
                    row_data["best_checkpoint"] = best_checkpoint.output_score_repr(
                        output_name=output_name
                    )
                if most_recent_checkpoint is not None:
                    row_data[
                        "most_recent_checkpoint"
                    ] = most_recent_checkpoint.output_score_repr(output_name)
                if worst_checkpoint is not None:
                    row_data["worst_checkpoint"] = worst_checkpoint.output_score_repr(
                        output_name=output_name
                    )
                input_specific_rows.append(row_data)
            if output_name == TASK_OUTPUT_KEY:
                per_input_filename = os.path.join(
                    self.data_dir, task_name, "per_input_breakdown.md"
                )
                output_table(
                    input_specific_rows,
                    title=f"Task success breakdown by input",
                    markdown_file=per_input_filename,
                    print_table=False,
                )
                abs_path = os.path.abspath(per_input_filename)
                file_link = f"[link=file://{abs_path}]per_input_breakdown.md[/link]"
                print(f"\nA per input breakdown can be found here: {file_link}")
            else:
                output_table(
                    input_specific_rows,
                    title=f"Intermediary state: {output_name} breakdown by input",
                )

    def cleanup_old_checkpoints(
        self,
        task_name: Optional[str] = None,
        ask_before_delete: bool = True,
        days_to_look_back: int = 4,
    ) -> None:
        task_name = self._get_recorded_task_name(task_name=task_name)

        tags = get_system_tags()
        if len(tags) == 0:
            return

        checkpoints = self.recorder.get_checkpoints_by_tags(task_name, tags)
        if len(checkpoints) == 0:
            print(f"No checkpoints found for tag combination: {tags}")
            return

        input_hash_to_checkpoints: dict[str, list[Checkpoint]] = defaultdict(list)
        time_since = datetime.datetime.utcnow() - datetime.timedelta(
            days=days_to_look_back
        )
        for checkpoint in checkpoints:
            if checkpoint.created_at > time_since:
                input_hash_to_checkpoints[checkpoint.get_input_hash()].append(
                    checkpoint
                )

        for input_hash in input_hash_to_checkpoints.keys():
            checkpoints = input_hash_to_checkpoints[input_hash]
            checkpoints.sort(key=lambda x: x.created_at, reverse=True)
            if len(checkpoints) == 1:
                continue

            num_deleted = 0
            if ask_before_delete:
                print(
                    f"\nFound {len(checkpoints)} checkpoints for input hash: {input_hash} for tags: {tags}"
                )
                val = input(
                    f"Enter 'Y' to delete {len(checkpoints) - 1} checkpoints and keep most recent checkpoint -- {checkpoints[0].checkpoint_id}\n"
                )
                if val != "Y":
                    print("Skipping deletion")
                    continue

            for checkpoint in checkpoints[1:]:
                self.recorder.delete_checkpoint(task_name, checkpoint.checkpoint_id)
                num_deleted += 1
            print(
                f"Deleted {num_deleted} checkpoints for input hash: {input_hash} for tags: {tags}"
            )


_default_evaluator = Evaluator()


set_config = _default_evaluator.set_config
record_task = _default_evaluator.record_task
record_output = _default_evaluator.record_output
record_input = _default_evaluator.record_input
get_eval_param = _default_evaluator.get_eval_param
get_eval_params = _default_evaluator.get_eval_params
rerun_recorded_examples = _default_evaluator.rerun_recorded_examples
rate_recorded_examples = _default_evaluator.rate_recorded_examples
evaluate_system = _default_evaluator.evaluate_system
delete_checkpoints_for_input_vars = _default_evaluator.delete_checkpoints_for_input_vars
calculate_system_health = _default_evaluator.calculate_system_health
record_intermediary_state = _default_evaluator.record_intermediary_state
start_recording_session = _default_evaluator.start_recording_session
default_evaluator = _default_evaluator
get_default_recorder = _default_evaluator.get_recorder
cleanup_old_checkpoints = _default_evaluator.cleanup_old_checkpoints
get_config_dict = _default_evaluator.get_config_dict


from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
import hashlib
import json
import yaml
import os
import pickle
from typing import Any, Optional, Protocol
from dataclasses import dataclass
import dataclasses
import requests

from .utils import LruCache
from .classes import MultiOutputFeedback, OutputFeedback, OutputScore, MultiOutputScores
import logging

logger = logging.getLogger(__name__)


@dataclass
class ComparisonResult:
    older_checkpoint_id: str
    newer_checkpoint_id: str
    feedback: MultiOutputFeedback

    def as_dict(self):
        return {
            "older_checkpoint_id": self.older_checkpoint_id,
            "newer_checkpoint_id": self.newer_checkpoint_id,
            "feedback": self.feedback.as_dict(),
        }

    @staticmethod
    def from_dict(data):
        return ComparisonResult(
            older_checkpoint_id=data["older_checkpoint_id"],
            newer_checkpoint_id=data["newer_checkpoint_id"],
            feedback=MultiOutputFeedback.from_dict(data["feedback"]),
        )


def get_input_hash(input_variables: dict[str, str]) -> str:
    sorted_input_vars = sorted(
        [
            (str(var_name), str(var_val))
            for var_name, var_val in input_variables.items()
        ],
        key=lambda x: x[0],
    )
    input_var_str = json.dumps(sorted_input_vars)
    return hashlib.sha256(input_var_str.encode("utf-8")).hexdigest()


@dataclass
class Checkpoint:
    checkpoint_id: str
    input_variables: dict[str, str] = dataclasses.field(default_factory=dict)
    eval_params: dict[str, Any] = dataclasses.field(default_factory=dict)
    intermediary_state: dict[str, str] = dataclasses.field(default_factory=dict)
    output: Optional[Any] = None
    feedback: MultiOutputFeedback = dataclasses.field(
        default_factory=MultiOutputFeedback
    )
    scores: MultiOutputScores = dataclasses.field(default_factory=MultiOutputScores)
    rerun_metadata: dict[str, str] = dataclasses.field(default_factory=dict)
    tags: list[str] = dataclasses.field(default_factory=list)

    def get_input_variables(self) -> dict[str, str]:
        return dict(self.input_variables)

    def get_input_hash(self) -> str:
        return get_input_hash(self.input_variables)

    def get_input_var_str(self) -> str:
        sorted_input_vars = sorted(
            [
                (str(var_name), str(var_val))
                for var_name, var_val in self.input_variables.items()
            ],
            key=lambda x: x[0],
        )
        msg = ""
        for var_name, var_val in sorted_input_vars:
            msg += f"{var_name}={var_val}\n"
        return msg

    def __str__(self) -> str:
        msg = f"Checkpoint: {self.checkpoint_id}\n"
        msg += self.get_input_var_str()
        return msg

    def as_dict(self) -> dict[str, Any]:
        checkpoint_dict = {
            "checkpoint_id": self.checkpoint_id,
            "input_variables": self.input_variables,
            "output": self.output,
        }
        if self.intermediary_state:
            checkpoint_dict["intermediary_state"] = self.intermediary_state
        if self.eval_params:
            checkpoint_dict["eval_params"] = self.eval_params
        if self.feedback is not None:
            checkpoint_dict["feedback"] = self.feedback
        if self.scores is not None and len(self.scores) > 0:
            checkpoint_dict["score"] = self.scores.as_dict()
        if self.rerun_metadata:
            checkpoint_dict["rerun_metadata"] = self.rerun_metadata

        return checkpoint_dict

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Checkpoint":
        return cls(
            checkpoint_id=data["checkpoint_id"],
            input_variables=data.get("input_variables", {}),
            eval_params=data.get("eval_params") or {},
            output=data.get("output"),
            feedback=MultiOutputFeedback.from_dict(data["feedback"])
            if data.get("feedback") is not None
            else {},
            scores=MultiOutputScores.from_dict(data.get("scores"))
            if data.get("scores") is not None
            else {},
            rerun_metadata=data.get("rerun_metadata") or {},
            intermediary_state=data.get("intermediary_state") or {},
        )

    @property
    def created_at(self) -> datetime:
        return datetime.fromisoformat(self.checkpoint_id)

    def output_score_repr(self, output_name: str) -> str:
        time_ago = datetime.now() - self.created_at
        if time_ago.days > 0:
            msg = f"@ {time_ago.days} days ago"
        elif time_ago.seconds > 3600:
            msg = f"@ {time_ago.seconds // 3600} hours ago"
        elif time_ago.seconds > 60:
            msg = f"@ {time_ago.seconds // 60} minutes ago"
        else:
            msg = f"@ {time_ago.seconds} seconds ago"
        msg += "  "

        if output_name in self.scores:
            output_score = self.scores[output_name]
            msg += f"score: {output_score.score:.2f}"
            #            if output_score.message:
            #                msg += f" ({output_score.message})"
            msg += "  "

        params_str = ", ".join(f"{k}={v}" for k, v in self.eval_params.items())
        if params_str:
            msg += f"({params_str})  "

        if self.output is not None and isinstance(self.output, str):
            msg += self.output[:140]
        return msg


class EvalRecorder(Protocol):
    def record_input_variable(
        self,
        task_name: str,
        checkpoint_id: str,
        variable_name: str,
        value: str,
    ) -> None:
        ...

    def record_output(
        self,
        task_name: str,
        checkpoint_id: str,
        output: str,
    ) -> None:
        ...

    def record_intermediary_state(
        self,
        task_name: str,
        checkpoint_id: str,
        state_name: str,
        state_value: str,
    ) -> None:
        ...

    def record_eval_params(
        self,
        task_name: str,
        checkpoint_id: str,
        eval_params: dict[str, Any],
        rerun_metadata: Optional[dict[str, Any]] = None,
    ) -> None:
        ...

    def record_output_feedback(
        self,
        task_name: str,
        checkpoint_id: str,
        feedback: MultiOutputFeedback,
    ) -> None:
        ...

    def record_output_scores(
        self,
        task_name: str,
        checkpoint_id: str,
        scores: MultiOutputScores,
    ) -> None:
        ...

    def record_comparison_result(
        self,
        task_name: str,
        input_hash: str,
        result: ComparisonResult,
    ) -> None:
        ...

    def add_checkpoint_tag(
        self,
        task_name: str,
        checkpoint_id: str,
        tag: str,
    ) -> None:
        ...

    def get_checkpoint(
        self, task_name: str, checkpoint_id: str
    ) -> Optional[Checkpoint]:
        ...

    def get_latest_checkpoints(
        self, task_name: str, input_hash: str, num_checkpoints: int = 2
    ) -> list[Checkpoint]:
        ...

    def get_input_hashes(self, task_name: str) -> list[str]:
        ...

    def get_task_names(self) -> list[str]:
        ...

    def get_comparison_result(
        self,
        task_name: str,
        older_checkpoint_id: str,
        newer_checkpoint_id: str,
    ) -> Optional[ComparisonResult]:
        ...

    def get_comparison_results_for_input_hash(
        self, task_name: str, input_hash: str, num_results: int = 3
    ) -> list[ComparisonResult]:
        ...

    def delete_checkpoints_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
    ) -> None:
        ...

    def get_checkpoints_by_tags(
        self, task_name: str, tags: list[str]
    ) -> list[Checkpoint]:
        ...

    def delete_checkpoint(self, task_name: str, checkpoint_id: str) -> None:
        ...


class ApiClientRecorder(EvalRecorder):
    def __init__(self, api_key: str, api_url: Optional[str] = None) -> None:
        self.api_key = api_key
        if api_url is None:
            self.url = "https://eyeball-dot-robust-ai.uc.r.appspot.com"
        else:
            self.url = api_url
        self.checkpoint_dicts: LruCache = LruCache(max_size=100)
        self.comparison_results_dict: LruCache = LruCache(max_size=100)

        self.pool = ThreadPoolExecutor(max_workers=1)

    def _get_headers(self) -> dict[str, str]:
        return {
            "Authorization": f"Bearer {self.api_key}",
        }

    def __getstate__(self):
        return {
            "api_key": self.api_key,
            "api_url": self.url,
            "checkpoints_dict": self.checkpoint_dicts,
            "comparison_reuslts_dict": self.comparison_results_dict
        }

    def __setstate__(self, state):
        self.api_key = state["api_key"]
        self.url = state["api_url"]
        self.checkpoint_dicts = state["checkpoints_dict"]
        self.comparison_results_dict = state["comparison_reuslts_dict"]


    def _record_checkpoint(
        self,
        task_name: str,
        checkpoint_id: str,
        prefixes: list[str],
        name: str,
        value: Any,
        flush: bool = False,
    ) -> dict[str, Any]:
        dict_key = f"{task_name},{checkpoint_id}"
        if dict_key not in self.checkpoint_dicts:
            checkpoint_dict: dict[str, Any] = {"checkpoint_id": checkpoint_id}
        else:
            checkpoint_dict = self.checkpoint_dicts[dict_key]

        current_dict = checkpoint_dict
        for prefix in prefixes:
            if prefix not in current_dict:
                current_dict[prefix] = {}
            current_dict = current_dict[prefix]
        current_dict[name] = value
        self.checkpoint_dicts[dict_key] = dict(checkpoint_dict)

        if flush:

            def _record():
                response = requests.post(
                    f"{self.url}/record_checkpoint",
                    json={
                        "task_name": task_name,
                        "checkpoint_id": checkpoint_id,
                        **checkpoint_dict,
                    },
                    headers=self._get_headers(),
                )
                if response.status_code != 200:
                    logger.error(
                        f"Failed to record checkpoint {checkpoint_id} -- {response.status_code}, {response.text}"
                    )
                else:
                    logger.debug(f"Recorded checkpoint {checkpoint_id}")

            self.pool.submit(_record)
        return checkpoint_dict

    def record_input_variable(
        self, task_name: str, checkpoint_id: str, variable_name: str, value: str
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=["input_variables"],
            name=variable_name,
            value=value,
        )

    def record_output(self, task_name: str, checkpoint_id: str, output: str) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="output",
            value=output,
            flush=True,
        )

    def record_intermediary_state(
        self, task_name: str, checkpoint_id: str, state_name: str, state_value: str
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=["intermediary_state"],
            name=state_name,
            value=state_value,
        )

    def record_eval_params(
        self,
        task_name: str,
        checkpoint_id: str,
        eval_params: dict[str, Any],
        rerun_metadata: Optional[dict[str, Any]] = None,
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="eval_params",
            value=eval_params,
        )
        if rerun_metadata is not None:
            self._record_checkpoint(
                task_name=task_name,
                checkpoint_id=checkpoint_id,
                prefixes=[],
                name="rerun_metadata",
                value=rerun_metadata,
            )

    def record_comparison_result(
        self, task_name: str, input_hash: str, result: ComparisonResult
    ) -> None:
        def _record():
            requests.post(
                f"{self.url}/record_comparison_result",
                json={
                    "task_name": task_name,
                    "input_hash": input_hash,
                    **result.as_dict(),
                },
                headers=self._get_headers(),
            )

        self.pool.submit(_record)
        key = f"{task_name},{result.older_checkpoint_id},{result.newer_checkpoint_id}"
        self.comparison_results_dict[key] = result

    def record_output_scores(
        self, task_name: str, checkpoint_id: str, scores: dict[str, OutputScore]
    ) -> None:
        requests.post(
            f"{self.url}/record_output_scores",
            json={
                "task_name": task_name,
                "checkpoint_id": checkpoint_id,
                "scores": {name: score.as_dict() for name, score in scores.items()},
            },
            headers=self._get_headers(),
        )

    def record_output_feedback(
        self, task_name: str, checkpoint_id: str, feedback: MultiOutputFeedback
    ) -> None:
        requests.post(
            f"{self.url}/record_output_feedback",
            json={
                "task_name": task_name,
                "checkpoint_id": checkpoint_id,
                "feedback": feedback.as_dict(),
            },
            headers=self._get_headers(),
        )

    def delete_checkpoints_for_input_hash(
        self, task_name: str, input_hash: str
    ) -> None:
        ...

    def get_checkpoint(
        self, task_name: str, checkpoint_id: str
    ) -> Optional[Checkpoint]:
        dict_key = f"{task_name},{checkpoint_id}"
        if dict_key in self.checkpoint_dicts:
            return Checkpoint(**(self.checkpoint_dicts[dict_key]))

        response = requests.get(
            f"{self.url}/get_checkpoint",
            params={
                "task_name": task_name,
                "checkpoint_id": checkpoint_id,
            },
            headers=self._get_headers(),
        )
        if response.status_code != 200:
            logger.debug(
                f"Failed to get checkpoint - {checkpoint_id}: {response.status_code}"
            )
            return None
        return Checkpoint.from_dict(response.json())

    def get_comparison_result(
        self, task_name: str, older_checkpoint_id: str, newer_checkpoint_id: str
    ) -> Optional[ComparisonResult]:
        key = f"{task_name},{older_checkpoint_id},{newer_checkpoint_id}"
        if key in self.comparison_results_dict:
            return self.comparison_results_dict[key]
        else:
            response = requests.get(
                f"{self.url}/get_comparison_result",
                params={
                    "task_name": task_name,
                    "older_checkpoint_id": older_checkpoint_id,
                    "newer_checkpoint_id": newer_checkpoint_id,
                },
                headers=self._get_headers(),
            )
            if response.status_code != 200:
                logger.debug(f"Failed to get comparison result: {response.status_code}")
                return None
            response_dict = response.json()
            if "success" in response_dict and not response_dict["success"]:
                return None
            return ComparisonResult.from_dict(response_dict)

    def get_comparison_results_for_input_hash(
        self, task_name: str, input_hash: str, num_results: int = 3
    ) -> list[ComparisonResult]:
        response = requests.get(
            f"{self.url}/get_comparison_results_for_input_hash",
            params={
                "task_name": task_name,
                "input_hash": input_hash,
                "num_results": num_results,
            },
            headers=self._get_headers(),
        )
        if response.status_code != 200:
            logger.debug(f"Failed to get comparison results: {response.status_code}")
            return []
        return [
            ComparisonResult.from_dict(data)
            for data in response.json()["comparison_results"]
        ]

    def get_input_hashes(self, task_name: str) -> list[str]:
        response = requests.get(
            f"{self.url}/get_input_hashes",
            params={
                "task_name": task_name,
            },
            headers=self._get_headers(),
        )
        if response.status_code != 200:
            logger.debug(f"Failed to get input hashes: {response.status_code}")
            return []
        return response.json()["input_hashes"]

    def get_latest_checkpoints(
        self, task_name: str, input_hash: str, num_checkpoints: int = 2
    ) -> list[Checkpoint]:
        response = requests.get(
            f"{self.url}/get_latest_checkpoints",
            params={
                "task_name": task_name,
                "input_hash": input_hash,
                "num_checkpoints": num_checkpoints,
            },
            headers=self._get_headers(),
        )
        if response.status_code != 200:
            logger.debug(f"Failed to get latest checkpoints: {response.status_code}")
            return []
        return [Checkpoint.from_dict(data) for data in response.json()["checkpoints"]]

    def get_task_names(self) -> list[str]:
        response = requests.get(
            f"{self.url}/get_task_names",
            headers=self._get_headers(),
        )
        if response.status_code != 200:
            logger.debug(f"Failed to get task names: {response.status_code}")
            return []
        return response.json()["task_names"]

    def add_checkpoint_tag(self, task_name: str, checkpoint_id: str, tag: str) -> None:
        return None

    def get_checkpoints_by_tags(
        self, task_name: str, tags: list[str]
    ) -> list[Checkpoint]:
        return []

    def delete_checkpoint(self, task_name: str, checkpoint_id: str) -> None:
        return None

@dataclass
class Task:
    name: str
    checkpoints: dict[str, Checkpoint] = dataclasses.field(default_factory=dict)
    input_hashes: dict[str, set[str]] = dataclasses.field(default_factory=dict)
    comparison_results: dict[str, ComparisonResult] = dataclasses.field(
        default_factory=dict
    )
    input_hash_to_comparison_results: dict[str, set[str]] = dataclasses.field(
        default_factory=dict
    )
    tags_to_checkpoints: dict[str, set[str]] = dataclasses.field(default_factory=dict)


class MemoryRecorder(EvalRecorder):
    def __init__(self) -> None:
        self.tasks: dict[str, Task] = {}

    def _fetch_or_create_checkpoint(
        self, task_name: str, checkpoint_id: str
    ) -> Checkpoint:
        if task_name not in self.tasks:
            task = Task(name=task_name)
            self.tasks[task_name] = task
        else:
            task = self.tasks[task_name]

        if checkpoint_id not in task.checkpoints:
            checkpoint = Checkpoint(
                checkpoint_id=checkpoint_id,
                input_variables={},
                eval_params={},
            )
            task.checkpoints[checkpoint_id] = checkpoint
            return checkpoint
        else:
            return task.checkpoints[checkpoint_id]

    def _save_input_hash(self, task_name: str, checkpoint: Checkpoint) -> None:
        # Remove the checkpoint from any previous input hashes
        task = self.tasks[task_name]
        checkpoint_id = checkpoint.checkpoint_id

        input_hashes_to_remove = []
        for input_hash in task.input_hashes:
            if checkpoint_id in task.input_hashes[input_hash]:
                task.input_hashes[input_hash].remove(checkpoint_id)
            if len(task.input_hashes[input_hash]) == 0:
                input_hashes_to_remove.append(input_hash)

        for input_hash in input_hashes_to_remove:
            del task.input_hashes[input_hash]

        # Add the checkpoint to the new input hash
        input_hash = checkpoint.get_input_hash()
        if input_hash not in task.input_hashes:
            task.input_hashes[input_hash] = set()
        task.input_hashes[input_hash].add(checkpoint_id)

    def record_input_variable(
        self,
        task_name: str,
        checkpoint_id: str,
        variable_name: str,
        value: str,
    ) -> None:
        checkpoint = self._fetch_or_create_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        checkpoint.input_variables[variable_name] = value
        self._save_input_hash(task_name, checkpoint)

    def record_output(self, task_name: str, checkpoint_id: str, output: str) -> None:
        checkpoint = self._fetch_or_create_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        checkpoint.output = output

    def record_intermediary_state(
        self, task_name: str, checkpoint_id: str, state_name: str, state_value: str
    ) -> None:
        checkpoint = self._fetch_or_create_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        checkpoint.intermediary_state[state_name] = state_value

    def get_latest_checkpoints(
        self, task_name: str, input_hash: str, num_checkpoints: int = 2
    ) -> list[Checkpoint]:
        if task_name not in self.tasks:
            return []
        task = self.tasks[task_name]
        if input_hash not in task.input_hashes:
            return []

        return [
            task.checkpoints[checkpoint_id]
            for checkpoint_id in sorted(task.input_hashes[input_hash])[
                -num_checkpoints:
            ]
        ]

    def get_checkpoint(
        self, task_name: str, checkpoint_id: str
    ) -> Optional[Checkpoint]:
        if task_name not in self.tasks:
            return None
        task = self.tasks[task_name]

        if checkpoint_id not in task.checkpoints:
            return None

        return task.checkpoints[checkpoint_id]

    def get_input_hashes(self, task_name: str) -> list[str]:
        task = self.tasks.get(task_name)
        if task is None:
            return []

        hashes_to_remove = []
        for input_hash in task.input_hashes:
            if len(task.input_hashes[input_hash]) == 0:
                hashes_to_remove.append(input_hash)

        for input_hash in hashes_to_remove:
            del task.input_hashes[input_hash]

        return list(task.input_hashes.keys())

    def record_eval_params(
        self,
        task_name: str,
        checkpoint_id: str,
        eval_params: dict[str, Any],
        rerun_metadata: Optional[dict[str, Any]] = None,
    ) -> None:
        checkpoint = self._fetch_or_create_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        checkpoint.eval_params.update(eval_params)
        if rerun_metadata is not None:
            checkpoint.rerun_metadata = rerun_metadata

    def record_output_feedback(
        self,
        task_name: str,
        checkpoint_id: str,
        feedback: MultiOutputFeedback,
    ) -> None:
        checkpoint = self._fetch_or_create_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        checkpoint.feedback = feedback

    def get_task_names(self) -> list[str]:
        return list(self.tasks.keys())

    def record_comparison_result(
        self,
        task_name: str,
        input_hash: str,
        result: ComparisonResult,
    ) -> None:
        task = self.tasks[task_name]
        key = f"{result.older_checkpoint_id},{result.newer_checkpoint_id}"
        task.comparison_results[key] = result

        if input_hash not in task.input_hash_to_comparison_results:
            task.input_hash_to_comparison_results[input_hash] = set()
        task.input_hash_to_comparison_results[input_hash].add(key)

    def record_output_scores(
        self,
        task_name: str,
        checkpoint_id: str,
        scores: MultiOutputScores,
    ) -> None:
        checkpoint = self._fetch_or_create_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        checkpoint.scores = scores

    def get_comparison_result(
        self,
        task_name: str,
        older_checkpoint_id: str,
        newer_checkpoint_id: str,
    ) -> Optional[ComparisonResult]:
        task = self.tasks[task_name]
        key = f"{older_checkpoint_id},{newer_checkpoint_id}"
        if key not in task.comparison_results:
            return None
        return task.comparison_results[key]

    def get_comparison_results_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
        num_results: int = 3,
    ) -> list[ComparisonResult]:
        task = self.tasks[task_name]
        if input_hash not in task.input_hash_to_comparison_results:
            return []
        comparison_results = task.input_hash_to_comparison_results[input_hash]
        return [task.comparison_results[key] for key in comparison_results][
            :num_results
        ]

    def delete_checkpoints_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
    ) -> None:
        task = self.tasks[task_name]
        if input_hash not in task.input_hashes:
            return
        for checkpoint_id in task.input_hashes[input_hash]:
            del task.checkpoints[checkpoint_id]
        del task.input_hashes[input_hash]

    def add_checkpoint_tag(self, task_name: str, checkpoint_id: str, tag: str) -> None:
        self.tasks[task_name].checkpoints[checkpoint_id].tags.append(tag)
        if tag not in self.tasks[task_name].tags_to_checkpoints:
            self.tasks[task_name].tags_to_checkpoints[tag] = set()

        self.tasks[task_name].tags_to_checkpoints[tag].add(checkpoint_id)

    def get_checkpoints_by_tags(
        self, task_name: str, tags: list[str]
    ) -> list[Checkpoint]:
        if task_name not in self.tasks:
            return []
        task = self.tasks[task_name]

        checkpoint_ids: Optional[set[str]] = None
        for tag in tags:
            if tag not in task.tags_to_checkpoints:
                return []

            if checkpoint_ids is None:
                checkpoint_ids = set(task.tags_to_checkpoints[tag])
            else:
                checkpoint_ids = checkpoint_ids.intersection(
                    task.tags_to_checkpoints[tag]
                )

            if len(checkpoint_ids) == 0:
                return []
        if checkpoint_ids is None:
            return []

        return [task.checkpoints[checkpoint_id] for checkpoint_id in checkpoint_ids]

    def delete_checkpoint(self, task_name: str, checkpoint_id: str) -> None:
        task = self.tasks[task_name]
        if checkpoint_id not in task.checkpoints:
            return
        checkpoint = task.checkpoints[checkpoint_id]
        input_hash = checkpoint.get_input_hash()
        if input_hash in task.input_hashes:
            task.input_hashes[input_hash].remove(checkpoint_id)
            if len(task.input_hashes[input_hash]) == 0:
                del task.input_hashes[input_hash]

        for tag in checkpoint.tags:
            task.tags_to_checkpoints[tag].remove(checkpoint_id)
            if len(task.tags_to_checkpoints[tag]) == 0:
                del task.tags_to_checkpoints[tag]

        del task.checkpoints[checkpoint_id]


class FileRecorder(EvalRecorder):
    def __init__(self, dir_path: str) -> None:
        self._dir_path = dir_path
        self.yaml_dicts: dict[str, dict[str, str]] = {}

    @property
    def dir_path(self) -> str:
        if not os.path.exists(self._dir_path):
            os.makedirs(self._dir_path)
        return self._dir_path

    def _edit_checkpoint(
        self, task_name: str, checkpoint_id: str, edit_dict: dict[str, Any]
    ) -> None:
        old_checkpoint = self.get_checkpoint(task_name, checkpoint_id)
        if old_checkpoint is None:
            return
        old_input_hash = old_checkpoint.get_input_hash()
        checkpoint_dict = old_checkpoint.as_dict()
        checkpoint_dict.update(edit_dict)
        new_checkpoint = Checkpoint.from_dict(checkpoint_dict)
        new_input_hash = new_checkpoint.get_input_hash()
        checkpoint_dict["input_hash"] = new_input_hash

        dir_name = os.path.join(self.dir_path, task_name, "checkpoints")
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        file_name = os.path.join(dir_name, f"{checkpoint_id}.yaml")
        yaml.dump(checkpoint_dict, open(file_name, "w+"))

        if old_input_hash != new_input_hash:
            inputs_dir_name = os.path.join(self.dir_path, task_name, "inputs")
            if not os.path.exists(inputs_dir_name):
                os.makedirs(inputs_dir_name)

            input_file_name = os.path.join(
                self.dir_path, task_name, "inputs", f"{new_input_hash}.yaml"
            )
            yaml.dump(new_checkpoint.input_variables, open(input_file_name, "w+"))
            try:
                os.remove(os.path.join(inputs_dir_name, f"{old_input_hash}.yaml"))
            except FileNotFoundError:
                pass

    def _record_checkpoint(
        self,
        task_name: str,
        checkpoint_id: str,
        prefixes: list[str],
        name: str,
        value: Any,
        flush: bool = False,
    ) -> dict[str, Any]:
        dir_name = os.path.join(self.dir_path, task_name, "checkpoints")
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        file_name = os.path.join(dir_name, f"{checkpoint_id}.yaml")

        key_name = f"{task_name},{checkpoint_id}"

        if key_name not in self.yaml_dicts:
            if os.path.exists(file_name):
                yaml_dict = yaml.load(open(file_name, "r"), Loader=yaml.FullLoader)
            else:
                yaml_dict = {}
        else:
            yaml_dict = self.yaml_dicts[key_name]

        current_dict = yaml_dict
        for prefix in prefixes:
            if prefix not in current_dict:
                current_dict[prefix] = {}
            current_dict = current_dict[prefix]
        current_dict[name] = value
        if flush:
            yaml.dump(yaml_dict, open(file_name, "w+"))
            try:
                del self.yaml_dicts[key_name]
            except KeyError:
                pass
        else:
            self.yaml_dicts[key_name] = dict(yaml_dict)
        return yaml_dict

    def record_input_variable(
        self, task_name: str, checkpoint_id: str, variable_name: str, value: str
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=["input_variables"],
            name=variable_name,
            value=value,
        )

    def record_output(
        self,
        task_name: str,
        checkpoint_id: str,
        output: str,
    ) -> None:
        yaml_dict = self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="output",
            value=output,
        )

        input_hash = get_input_hash(yaml_dict["input_variables"])
        inputs_dir_name = os.path.join(self.dir_path, task_name, "inputs")
        if not os.path.exists(inputs_dir_name):
            os.makedirs(inputs_dir_name)

        input_file_name = os.path.join(
            self.dir_path, task_name, "inputs", f"{input_hash}.yaml"
        )
        yaml.dump(yaml_dict["input_variables"], open(input_file_name, "w+"))
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="input_hash",
            value=input_hash,
            flush=True,
        )

    def record_intermediary_state(
        self,
        task_name: str,
        checkpoint_id: str,
        state_name: str,
        state_value: str,
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=["intermediary_state"],
            name=state_name,
            value=state_value,
        )

    def record_eval_params(
        self,
        task_name: str,
        checkpoint_id: str,
        eval_params: dict[str, Any],
        rerun_metadata: Optional[dict[str, Any]] = None,
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="eval_params",
            value=eval_params,
        )
        if rerun_metadata is not None:
            self._record_checkpoint(
                task_name=task_name,
                checkpoint_id=checkpoint_id,
                prefixes=[],
                name="rerun_metadata",
                value=rerun_metadata,
            )

    def record_output_feedback(
        self,
        task_name: str,
        checkpoint_id: str,
        feedback: MultiOutputFeedback,
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="feedback",
            value=feedback.as_dict(),
            flush=True,
        )

    def record_output_scores(
        self, task_name: str, checkpoint_id: str, scores: MultiOutputScores
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="scores",
            value=scores.as_dict(),
            flush=True,
        )

    def _checkpoint_from_yaml_dict(
        self, yaml_dict: dict[str, Any], checkpoint_id: str
    ) -> Checkpoint:
        checkpoint = Checkpoint(checkpoint_id=checkpoint_id)
        checkpoint.input_variables = yaml_dict.get("input_variables", {})
        checkpoint.eval_params = yaml_dict.get("eval_params", {})
        checkpoint.output = yaml_dict.get("output")
        checkpoint.feedback = (
            MultiOutputFeedback.from_dict(yaml_dict.get("feedback"))
            if "feedback" in yaml_dict
            else None
        )
        checkpoint.scores = (
            MultiOutputScores.from_dict(yaml_dict.get("scores"))
            if "scores" in yaml_dict
            else None
        )
        checkpoint.rerun_metadata = yaml_dict.get("rerun_metadata") or {}
        return checkpoint

    def get_checkpoint(
        self, task_name: str, checkpoint_id: str
    ) -> Optional[Checkpoint]:
        key_name = f"{task_name},{checkpoint_id}"
        if key_name in self.yaml_dicts:
            return Checkpoint.from_dict(self.yaml_dicts[key_name] | {"checkpoint_id": checkpoint_id})

        file_name = os.path.join(
            self.dir_path, task_name, "checkpoints", f"{checkpoint_id}.yaml"
        )
        if not os.path.exists(file_name):
            return None
        yaml_dict = yaml.load(open(file_name, "r"), Loader=yaml.FullLoader)
        yaml_dict["checkpoint_id"] = checkpoint_id
        return Checkpoint.from_dict(yaml_dict)

    def _read_checkpoints(self, task_name: str) -> dict[str, list[Checkpoint]]:
        input_hashes_to_checkpoints: dict[str, list[Checkpoint]] = {}
        dir_name = os.path.join(self.dir_path, task_name, "checkpoints")
        for file_name in os.listdir(dir_name):
            if file_name.endswith(".yaml"):
                file_path = os.path.join(dir_name, file_name)
                yaml_dict = yaml.load(open(file_path, "r"), Loader=yaml.FullLoader)
                if input_hash := yaml_dict.get("input_hash"):
                    if input_hash not in input_hashes_to_checkpoints:
                        input_hashes_to_checkpoints[input_hash] = []
                    yaml_dict["checkpoint_id"] = file_name[:-5]
                    input_hashes_to_checkpoints[input_hash].append(
                        Checkpoint.from_dict(yaml_dict)
                    )

        for input_hash in input_hashes_to_checkpoints:
            input_hashes_to_checkpoints[input_hash].sort(
                reverse=True, key=lambda x: x.checkpoint_id
            )
        return input_hashes_to_checkpoints

    def get_latest_checkpoints(
        self, task_name: str, input_hash: str, num_checkpoints: int = 2
    ) -> list[Checkpoint]:
        input_hashes_to_checkpoints = self._read_checkpoints(task_name)
        if input_hash not in input_hashes_to_checkpoints:
            return []
        return input_hashes_to_checkpoints[input_hash][:num_checkpoints]

    def get_input_hashes(self, task_name: str) -> list[str]:
        input_hashes = []
        for file_name in os.listdir(os.path.join(self.dir_path, task_name, "inputs")):
            if file_name.endswith(".yaml"):
                input_hashes.append(file_name[:-5])
        return input_hashes

    def get_task_names(self) -> list[str]:
        task_names = []
        for dir_name in os.listdir(self.dir_path):
            if os.path.isdir(os.path.join(self.dir_path, dir_name)):
                task_names.append(dir_name)
        return task_names

    def record_comparison_result(
        self,
        task_name: str,
        input_hash: str,
        result: ComparisonResult,
    ) -> None:
        comparison_dir = os.path.join(self.dir_path, task_name, "comparison_results")
        if not os.path.exists(comparison_dir):
            os.makedirs(comparison_dir)

        file_name = f"{input_hash}_{result.older_checkpoint_id}_{result.newer_checkpoint_id}.yaml"
        file_path = os.path.join(comparison_dir, file_name)
        yaml.dump(
            result.feedback.as_dict(),
            open(file_path, "w+"),
        )

    def get_comparison_result(
        self,
        task_name: str,
        older_checkpoint_id: str,
        newer_checkpoint_id: str,
    ) -> Optional[ComparisonResult]:
        comparison_dir = os.path.join(self.dir_path, task_name, "comparison_results")
        if not os.path.exists(comparison_dir):
            return None

        for file_name in os.listdir(comparison_dir):
            splits = os.path.splitext(file_name)[0].split("_")
            if len(splits) != 3:
                continue

            if splits[1] != older_checkpoint_id:
                continue

            if splits[2] != newer_checkpoint_id:
                continue

            if file_name.endswith(".yaml"):
                file_path = os.path.join(comparison_dir, file_name)
                yaml_dict = yaml.load(open(file_path, "r"), Loader=yaml.FullLoader)
                return ComparisonResult(
                    older_checkpoint_id=older_checkpoint_id,
                    newer_checkpoint_id=newer_checkpoint_id,
                    feedback=MultiOutputFeedback.from_dict(yaml_dict),
                )
        return None

    def get_comparison_results_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
        num_results: int = 3,
    ) -> list[ComparisonResult]:
        comparison_dir = os.path.join(self.dir_path, task_name, "comparison_results")
        if not os.path.exists(comparison_dir):
            return []

        results: list[ComparisonResult] = []
        for file_name in os.listdir(comparison_dir):
            splits = os.path.splitext(file_name)[0].split("_")
            if len(splits) != 3:
                continue

            if splits[0] != input_hash:
                continue

            if file_name.endswith(".yaml"):
                file_path = os.path.join(comparison_dir, file_name)
                yaml_dict = yaml.load(open(file_path, "r"), Loader=yaml.FullLoader)
                results.append(
                    ComparisonResult(
                        older_checkpoint_id=splits[1],
                        newer_checkpoint_id=splits[2],
                        feedback=MultiOutputFeedback.from_dict(yaml_dict),
                    )
                )
        return sorted(results, key=lambda x: x.newer_checkpoint_id, reverse=True)[
            :num_results
        ]

    def delete_checkpoints_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
    ) -> None:
        ...

    def add_checkpoint_tag(self, task_name: str, checkpoint_id: str, tag: str) -> None:
        checkpoint = self.get_checkpoint(task_name, checkpoint_id)
        if checkpoint is None:
            tags = [tag]
        else:
            tags = checkpoint.tags
            if tag not in tags:
                tags.append(tag)

        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="tags",
            value=tags,
        )

    def get_checkpoints_by_tags(
        self, task_name: str, tags: list[str]
    ) -> list[Checkpoint]:
        checkpoints = []
        for file_name in os.listdir(
            os.path.join(self.dir_path, task_name, "checkpoints")
        ):
            if file_name.endswith(".yaml"):
                file_path = os.path.join(
                    self.dir_path, task_name, "checkpoints", file_name
                )
                yaml_dict = yaml.load(open(file_path, "r"), Loader=yaml.FullLoader)
                if "tags" in yaml_dict:
                    for tag in tags:
                        if tag not in yaml_dict["tags"]:
                            continue
                    yaml_dict["checkpoint_id"] = file_name[:-5]
                    checkpoints.append(Checkpoint.from_dict(yaml_dict))
        return checkpoints

    def delete_checkpoint(self, task_name: str, checkpoint_id: str) -> None:
        dir_name = os.path.join(self.dir_path, task_name, "checkpoints")
        if not os.path.exists(dir_name):
            return
        file_name = os.path.join(dir_name, f"{checkpoint_id}.yaml")
        if not os.path.exists(file_name):
            return
        os.remove(file_name)


class DiskRecorder(EvalRecorder):
    # TODO: improve this by using rocksdb etc. vs a stupid pickle file
    def __init__(self, dir_path: str) -> None:
        self.file_name = os.path.join(dir_path, "eyeball_checkpoints.pkl")
        if os.path.exists(self.file_name):
            self.memory_recorder = pickle.load(open(self.file_name, "rb"))
        else:
            self.memory_recorder = MemoryRecorder()

    def record_input_variable(
        self,
        task_name: str,
        checkpoint_id: str,
        variable_name: str,
        value: str,
    ) -> None:
        self.memory_recorder.record_input_variable(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            variable_name=variable_name,
            value=value,
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def record_intermediary_state(
        self,
        task_name: str,
        checkpoint_id: str,
        state_name: str,
        state_value: str,
    ) -> None:
        self.memory_recorder.record_intermediary_state(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            state_name=state_name,
            state_value=state_value,
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def record_eval_params(
        self,
        task_name: str,
        checkpoint_id: str,
        eval_params: dict[str, Any],
        rerun_metadata: Optional[dict[str, Any]] = None,
    ) -> None:
        self.memory_recorder.record_eval_params(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            eval_params=eval_params,
            rerun_metadata=rerun_metadata,
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def record_output_feedback(
        self,
        task_name: str,
        checkpoint_id: str,
        feedback: MultiOutputFeedback,
    ) -> None:
        self.memory_recorder.record_output_feedback(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            feedback=feedback,
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def get_checkpoint(
        self, task_name: str, checkpoint_id: str
    ) -> Optional[Checkpoint]:
        return self.memory_recorder.get_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )

    def get_latest_checkpoints(
        self, task_name: str, input_hash: str, num_checkpoints: int = 2
    ) -> list[Checkpoint]:
        return self.memory_recorder.get_latest_checkpoints(
            task_name=task_name, input_hash=input_hash, num_checkpoints=num_checkpoints
        )

    def get_input_hashes(self, task_name: str) -> list[str]:
        return self.memory_recorder.get_input_hashes(task_name)

    def get_task_names(self) -> list[str]:
        return self.memory_recorder.get_task_names()

    def record_comparison_result(
        self,
        task_name: str,
        input_hash: str,
        result: ComparisonResult,
    ) -> None:
        self.memory_recorder.record_comparison_result(
            task_name=task_name, input_hash=input_hash, result=result
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def record_output_scores(
        self,
        task_name: str,
        checkpoint_id: str,
        scores: MultiOutputScores,
    ) -> None:
        self.memory_recorder.record_output_scores(
            task_name=task_name, checkpoint_id=checkpoint_id, scores=scores
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def get_comparison_result(
        self,
        task_name: str,
        older_checkpoint_id: str,
        newer_checkpoint_id: str,
    ) -> Optional[ComparisonResult]:
        return self.memory_recorder.get_comparison_result(
            task_name=task_name,
            older_checkpoint_id=older_checkpoint_id,
            newer_checkpoint_id=newer_checkpoint_id,
        )

    def get_comparison_results_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
        num_results: int = 3,
    ) -> list[ComparisonResult]:
        return self.memory_recorder.get_comparison_results_for_input_hash(
            task_name=task_name, input_hash=input_hash, num_results=num_results
        )

    def record_output(
        self,
        task_name: str,
        checkpoint_id: str,
        output: str,
    ) -> None:
        self.memory_recorder.record_output(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            output=output,
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def delete_checkpoints_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
    ) -> None:
        self.memory_recorder.delete_checkpoints_for_input_hash(
            task_name=task_name, input_hash=input_hash
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def add_checkpoint_tag(self, task_name: str, checkpoint_id: str, tag: str) -> None:
        self.memory_recorder.add_checkpoint_tag(
            task_name=task_name, checkpoint_id=checkpoint_id, tag=tag
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def get_checkpoints_by_tags(self, task_name: str, tags: list[str]) -> list[Checkpoint]:
        return self.memory_recorder.get_checkpoints_by_tags(
            task_name=task_name, tags=tags
        )

    def delete_checkpoint(self, task_name: str, checkpoint_id: str) -> None:
        self.memory_recorder.delete_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))



from dataclasses import dataclass, asdict
import json
import openai
import os
import tiktoken
from typing import Optional, Any
from .classes import (
    FeedbackResult,
    MultiOutputFeedback,
    OutputFeedback,
    OutputScore,
    TASK_OUTPUT_KEY,
)


def output_feedback_from_scores(
    older_scores: dict[str, OutputScore], newer_scores: dict[str, OutputScore]
) -> MultiOutputFeedback:
    feedback = {}
    for output_name, older_score in older_scores.items():
        newer_score = newer_scores.get(output_name)
        if newer_scores is None:
            continue

        if older_score == newer_score:
            feedback[output_name] = OutputFeedback(
                FeedbackResult.NEUTRAL,
                f"Score {older_scores} is equal to {newer_scores}",
            )
        elif older_score > newer_score:
            feedback[output_name] = OutputFeedback(
                FeedbackResult.NEGATIVE,
                f"Score {newer_scores} is worse than {newer_scores}",
            )
        else:
            feedback[output_name] = OutputFeedback(
                FeedbackResult.POSITIVE,
                f"Score {newer_scores} is better than {older_scores}",
            )
    return MultiOutputFeedback(feedback)


def _execute_summarizer(
    input_variables: dict[str, str],
    content: str,
) -> str:
    @dataclass
    class SummaryRequest:
        inputs: dict[str, str]
        content: str

    system_msg = "You are a summarizer that summarizes content within the context of the input parameters. You provide a detailed summary that does not leave out any pieces of information that could be used in the context of the input. If you are unsure, you include the content in your summary."
    summary_request = SummaryRequest(inputs=input_variables, content=content)
    user_msg = f"""{json.dumps(asdict(summary_request))}
    Tip: Be sure to leave in all the pieces of information that could be used in the context of the input. If you are unsure, you include the content in your summary.
    """

    response = openai.ChatCompletion.create(  # type: ignore
        model="gpt-3.5-turbo-16k",
        temperature=0.1,
        messages=[
            {"role": "system", "content": system_msg},
            {"role": "user", "content": user_msg},
        ],
    )["choices"][0]["message"]["content"]
    return response


def _execute_comparator(
    objective: str,
    input_variables: dict[str, str],
    older_checkpoint_response: str,
    newer_checkpoint_response: str,
) -> OutputFeedback:
    @dataclass
    class CompareRequest:
        objective: str
        inputs: dict[str, str]
        responses: list[dict[str, str]]

    system_msg = "You are an evaluator trying to grade the response of two agents based on provided JSON data. Keeping the objectives and the inputs in mind, decide which response is better and provide a reason. You always use the function provided."
    responses = [
        {
            "name": "Andrew",
            "response": older_checkpoint_response,
        },
        {
            "name": "Barry",
            "response": newer_checkpoint_response,
        },
    ]
    llm_request = CompareRequest(
        objective=objective, inputs=input_variables, responses=responses
    )
    user_msg = f"""{json.dumps(asdict(llm_request))}

    Given the above objective, inputs and responses, report your decision on the best response along with the reasoning. Think step by step.
    """
    functions = [
        {
            "name": "report_decision",
            "description": "report the results of the evaluation",
            "parameters": {
                "type": "object",
                "properties": {
                    "decision": {
                        "type": "string",
                        "enum": ["Andrew", "Barry", "Same"],
                        "description": "The name of the agent with the best response. 'Same' if both are equally good.",
                    },
                    "reason": {
                        "type": "string",
                        "description": "The reason for the decision.",
                    },
                },
                "required": ["decision", "reason"],
            },
        }
    ]

    response = openai.ChatCompletion.create(  # type: ignore
        model="gpt-4",
        temperature=0.1,
        messages=[
            {"role": "system", "content": system_msg},
            {"role": "user", "content": user_msg},
        ],
        functions=functions,
        function_call={"name": "report_decision"},
    )["choices"][0]["message"]

    assert response["content"] is None
    assert response["function_call"]["name"] == "report_decision"

    decision = json.loads(response["function_call"]["arguments"])["decision"]
    reason = json.loads(response["function_call"]["arguments"])["reason"]
    if "Andrew" in decision:
        return OutputFeedback(FeedbackResult.NEGATIVE, reason)
    elif "Barry" in decision:
        return OutputFeedback(FeedbackResult.POSITIVE, reason)
    else:
        return OutputFeedback(FeedbackResult.NEUTRAL, reason)


def _execute_llm_operation(
    objective: str,
    input_variables: dict[str, str],
    older_checkpoint_response: str,
    newer_checkpoint_response: str,
) -> OutputFeedback:
    token_encoding = tiktoken.get_encoding("cl100k_base")
    token_count = len(
        token_encoding.encode(older_checkpoint_response + newer_checkpoint_response)
    )
    COMPARISON_MAX_TOKEN_COUNT = int(os.environ.get("COMPARISON_MAX_TOKEN_COUNT", 6000))

    if token_count > COMPARISON_MAX_TOKEN_COUNT:
        older_checkpoint_response = _execute_summarizer(
            input_variables, older_checkpoint_response
        )
        newer_checkpoint_response = _execute_summarizer(
            input_variables, newer_checkpoint_response
        )

    return _execute_comparator(
        objective=objective,
        input_variables=input_variables,
        older_checkpoint_response=older_checkpoint_response,
        newer_checkpoint_response=newer_checkpoint_response,
    )


def model_graded_comparator(
    task_objective: str,
    input_variables: dict[str, str],
    older_checkpoint_output: str,
    newer_checkpoint_output: str,
    objectives_intermediary_state: Optional[dict[str, str]] = None,
    older_checkpoint_intermediary_state: Optional[dict[str, str]] = None,
    newer_checkpoint_intermediary_state: Optional[dict[str, str]] = None,
) -> MultiOutputFeedback:
    feedback = {}

    if older_checkpoint_intermediary_state and newer_checkpoint_intermediary_state:
        for key, value in older_checkpoint_intermediary_state.items():
            older_int_state = value
            newer_int_state = newer_checkpoint_intermediary_state[key]
            if objectives_intermediary_state and key in objectives_intermediary_state:
                feedback[key] = _execute_llm_operation(
                    objective=objectives_intermediary_state[key],
                    input_variables=input_variables,
                    older_checkpoint_response=older_int_state,
                    newer_checkpoint_response=newer_int_state,
                )
            else:
                print(f"Warning: No objective found for {key} in intermediary state")

    feedback[TASK_OUTPUT_KEY] = _execute_comparator(
        objective=task_objective,
        input_variables=input_variables,
        older_checkpoint_response=older_checkpoint_output,
        newer_checkpoint_response=newer_checkpoint_output,
    )

    return MultiOutputFeedback(feedback)



LLM_PRICING = {
    "gpt-3.5-turbo": (0.002 / 1000.0, 0.002 / 1000.0),
    "gpt-4": (0.03 / 1000.0, 0.06 / 1000.0),
}


def calculate_cost(model: str, input_tokens: int, output_tokens: int) -> float:
    """Calculate the cost of an LLM request"""
    input_token_price, output_token_price = LLM_PRICING.get(model, (0, 0))
    total_cost = (input_token_price * input_tokens) + \
        (output_token_price * output_tokens)
    return total_cost
